<div id="simplyprint_plugin_navbar">
    <div class="nav navbar-text">
        <a href="https://simplyprint.io/panel" target="_blank"><img alt="SimplyPrint logo (all rights reserved)" src="plugin/SimplyPrint/static/img/sp_logo.png" title="SimplyPrint active" style="width: 28px;"></a>

        <strong>
            <span id="simplyprint_version_wrapper">v{{ plugin_SimplyPrint_version }}</span>
        </strong>

        <div id="simplyprint_loading_info" data-bind="visible: loading">
            <i class="fa fa-spinner fa-spin" aria-hidden="true"></i>
        </div>
        <div id="simplyprint_is_set_up" data-bind="visible: settingsViewModel.settings.plugins.SimplyPrint.is_set_up">
            <!-- ko ifnot: settingsViewModel.settings.plugins.SimplyPrint.printer_name() === 'unset' || !settingsViewModel.settings.plugins.SimplyPrint.printer_name().length -->
            <span id="simplyprint_printer_name" data-bind="text: settingsViewModel.settings.plugins.SimplyPrint.printer_name"></span>
            <!-- /ko -->
            <!-- ko if: settingsViewModel.settings.plugins.SimplyPrint.printer_name() === 'unset' || !settingsViewModel.settings.plugins.SimplyPrint.printer_name().length -->
            <span id="simplyprint_printer_name">...</span>
            <!-- /ko -->
        </div>
        <div id="simplyprint_not_set_up" data-bind="visible: !settingsViewModel.settings.plugins.SimplyPrint.is_set_up()">
            <i class="fa fa-times" style="color:red;" aria-hidden="true"></i>
            Setup code:
            <div
                id="simplyprint_short_setup_code"
                data-bind="text: settingsViewModel.settings.plugins.SimplyPrint.temp_short_setup_id, css: {'splcd dot-flashing' : loading()}"
            >
            </div>
        </div>
    </div>
</div>